from awacs.aws import (
    Action,
    Allow,
    Bool,
    Condition,
    Policy,
    Principal,
    AWSPrincipal,
    Statement,
    StringEquals,
)
import awacs.logs
import awacs.s3
import awacs.kms
from awacs import sts
from stacker.blueprints.base import Blueprint
from troposphere import (
    iam,
    kms,
    s3,
    Equals,
    GetAtt,
    If,
    Join,
    Not,
    Or,
    Output,
    Ref,
)

from troposphere import Condition as TropoCondition

BUCKET = 'S3Bucket'
IAM_ROLE = 'IAMRole'
ROLE_POLICY = 'RolePolicy'
FIREHOSE_WRITE_POLICY = 'FirehoseWriteAccess'
LOGS_POLICY = 'LogsPolicy'
S3_WRITE_POLICY = 'S3WriteAccess'
LOGS_WRITE_POLICY = 'LogsWriteAccess'
KMS_KEY = "EncryptionKey"


class FirehoseAction(Action):
    def __init__(self, action=None):
        self.prefix = "firehose"
        self.action = action


def s3_arn(bucket):
    return Join('', ['arn:aws:s3:::', bucket])


def logs_policy():
    statements = [
        Statement(
            Effect=Allow,
            Action=[
                awacs.logs.CreateLogStream,
                awacs.logs.CreateLogGroup,
            ],
            Resource=['*'],
        ),
    ]
    return Policy(Statement=statements)


def firehose_write_policy():
    statements = [
        Statement(
            Effect=Allow,
            Action=[
                FirehoseAction("CreateDeliveryStream"),
                FirehoseAction("DeleteDeliveryStream"),
                FirehoseAction("DescribeDeliveryStream"),
                FirehoseAction("PutRecord"),
                FirehoseAction("PutRecordBatch"),
            ],
            Resource=['*'],
        ),
    ]
    return Policy(Statement=statements)


def logs_write_policy():
    statements = [
        Statement(
            Effect=Allow,
            Action=[
                awacs.logs.PutLogEvents,
            ],
            Resource=['*'],
        ),
    ]
    return Policy(Statement=statements)


def s3_write_policy(bucket):
    statements = [
        Statement(
            Effect=Allow,
            Action=[
                awacs.s3.AbortMultipartUpload,
                awacs.s3.GetBucketLocation,
                awacs.s3.GetObject,
                awacs.s3.ListBucket,
                awacs.s3.ListBucketMultipartUploads,
                awacs.s3.PutObject,
            ],
            Resource=[
                s3_arn(bucket),
                s3_arn(Join("/", [bucket, "*"]))
            ],
        ),
    ]
    return Policy(Statement=statements)


def kms_key_policy(key_use_arns, key_admin_arns):
    """ Creates a key policy for use of a KMS Key.

    key_use_arns is a list of arns that should have access to use the KMS
    key.
    """

    root_arn = Join(":", ["arn:aws:iam:", Ref("AWS::AccountId"), "root"])

    statements = []
    statements.append(
        Statement(
            Sid="Enable IAM User Permissions",
            Effect=Allow,
            Principal=AWSPrincipal(root_arn),
            Action=[
                Action("kms", "*"),
            ],
            Resource=["*"]
        )
    )
    statements.append(
        Statement(
            Sid="Allow use of the key",
            Effect=Allow,
            Principal=AWSPrincipal(key_use_arns),
            Action=[
                awacs.kms.Encrypt,
                awacs.kms.Decrypt,
                awacs.kms.ReEncrypt,
                awacs.kms.GenerateDataKey,
                awacs.kms.GenerateDataKeyWithoutPlaintext,
                awacs.kms.DescribeKey,
            ],
            Resource=["*"]
        )
    )
    statements.append(
        Statement(
            Sid="Allow attachment of persistent resources",
            Effect=Allow,
            Principal=AWSPrincipal(key_use_arns),
            Action=[
                awacs.kms.CreateGrant,
                awacs.kms.ListGrants,
                awacs.kms.RevokeGrant,
            ],
            Resource=["*"],
            Condition=Condition(Bool("kms:GrantIsForAWSResource", True))
        )
    )
    statements.append(
        Statement(
            Sid="Allow access for Key Administrators",
            Effect=Allow,
            Principal=AWSPrincipal(key_admin_arns),
            Action=[
                Action("kms", "Create*"),
                Action("kms", "Describe*"),
                Action("kms", "Enable*"),
                Action("kms", "List*"),
                Action("kms", "Put*"),
                Action("kms", "Update*"),
                Action("kms", "Revoke*"),
                Action("kms", "Disable*"),
                Action("kms", "Get*"),
                Action("kms", "Delete*"),
                Action("kms", "ScheduleKeyDeletion"),
                Action("kms", "CancelKeyDeletion"),
            ],
            Resource=["*"],
        )
    )

    return Policy(Version="2012-10-17", Id="key-default-1",
                  Statement=statements)


class Firehose(Blueprint):

    PARAMETERS = {
        "RoleNames": {
            "type": "CommaDelimitedList",
            "description": "A list of role names that should have access to "
                           "write to the firehose stream.",
            "default": "",
        },
        "GroupNames": {
            "type": "CommaDelimitedList",
            "description": "A list of group names that should have access to "
                           "write to the firehose stream.",
            "default": "",
        },
        "UserNames": {
            "type": "CommaDelimitedList",
            "description": "A list of user names that should have access to "
                           "write to the firehose stream.",
            "default": "",
        },
        "BucketName": {
            "type": "String",
            "description": "Name for the S3 Bucket",
        },
        "EncryptS3Bucket": {
            "type": "String",
            "description": "If set to true, a KMS key will be created to use "
                           "for encrypting the S3 Bucket's contents. If set "
                           "to false, no encryption will occur. Default: true",
            "allowed_values": ["true", "false"],
            "default": "true",
        },
        "EnableKeyRotation": {
            "type": "String",
            "description": "Whether to enable key rotation on the KMS key "
                           "generated if EncryptS3Bucket is set to true. "
                           "Default: true",
            "allowed_values": ["true", "false"],
            "default": "true",
        },
    }

    LOCAL_PARAMETERS = {
        # A list of ARNs that need access to the KMS key created with
        # EncryptS3Bucket
        "KeyUseArns": {
            "type": list,
            "default": [],
        },
        # A list of ARNs that need access to admin the KMS key created with
        # EncryptS3Bucket
        "KeyAdminArns": {
            "type": list,
            "default": [],
        }
    }

    def create_kms_key(self):
        t = self.template
        t.add_condition(
            "EncryptS3Bucket",
            Not(Equals(Ref("EncryptS3Bucket"), "false"))
        )

        key_description = Join(
            "",
            [
                "S3 Bucket kms encryption key for stack ",
                Ref("AWS::StackName")
            ]
        )

        key_use_arns = self.local_parameters["KeyUseArns"]
        # auto add the created IAM Role
        key_use_arns.append(GetAtt(IAM_ROLE, "Arn"))

        key_admin_arns = self.local_parameters["KeyAdminArns"]

        t.add_resource(
            kms.Key(
                KMS_KEY,
                Description=key_description,
                Enabled=True,
                EnableKeyRotation=Ref("EnableKeyRotation"),
                KeyPolicy=kms_key_policy(key_use_arns, key_admin_arns),
                Condition="EncryptS3Bucket"
            )
        )
        key_arn = Join(
            "",
            [
                "arn:aws:kms:",
                Ref("AWS::Region"),
                ":",
                Ref("AWS::AccountId"),
                ":key/",
                Ref(KMS_KEY)
            ]
        )
        t.add_output(Output("KmsKeyArn", Value=key_arn))

    def create_bucket(self):
        t = self.template
        t.add_resource(
            s3.Bucket(
                BUCKET,
                BucketName=Ref('BucketName'),
            )
        )
        t.add_output(Output('Bucket', Value=Ref(BUCKET)))

    def generate_iam_policies(self):
        ns = self.context.namespace
        name_prefix = "%s-%s" % (ns, self.name)
        s3_policy = iam.Policy(
            S3_WRITE_POLICY,
            PolicyName='{}-s3-write'.format(name_prefix),
            PolicyDocument=s3_write_policy(Ref('BucketName')),
        )
        logs_policy = iam.Policy(
            LOGS_WRITE_POLICY,
            PolicyName='{}-logs-write'.format(name_prefix),
            PolicyDocument=logs_write_policy(),
        )
        return [s3_policy, logs_policy]

    def create_role(self):
        t = self.template

        statements = [
            Statement(
                Principal=Principal('Service', ['firehose.amazonaws.com']),
                Effect=Allow,
                Action=[sts.AssumeRole],
                Condition=Condition(
                    StringEquals('sts:ExternalId', Ref('AWS::AccountId')),
                ),
            ),
        ]
        firehose_role_policy = Policy(Statement=statements)
        t.add_resource(
            iam.Role(
                IAM_ROLE,
                AssumeRolePolicyDocument=firehose_role_policy,
                Path='/',
                Policies=self.generate_iam_policies(),
            ),
        )
        t.add_output(Output('Role', Value=Ref(IAM_ROLE)))
        t.add_output(Output('RoleArn', Value=GetAtt(IAM_ROLE, 'Arn')))

    def create_policy(self):
        ns = self.context.namespace
        name_prefix = "%s-%s" % (ns, self.name)
        t = self.template

        t.add_condition(
            'ExternalRoles',
            Not(Equals(Join(",", Ref('RoleNames')), '')),
        )
        t.add_condition(
            'ExternalGroups',
            Not(Equals(Join(",", Ref('GroupNames')), '')),
        )
        t.add_condition(
            'ExternalUsers',
            Not(Equals(Join(",", Ref('UserNames')), '')),
        )

        t.add_condition(
            'CreatePolicy',
            Or(
                TropoCondition("ExternalRoles"),
                TropoCondition("ExternalGroups"),
                TropoCondition("ExternalUsers"),
            )
        )

        t.add_resource(
            iam.PolicyType(
                FIREHOSE_WRITE_POLICY,
                PolicyName='{}-firehose'.format(name_prefix),
                PolicyDocument=firehose_write_policy(),
                Roles=If("ExternalRoles",
                         Ref("RoleNames"),
                         Ref("AWS::NoValue")),
                Groups=If("ExternalGroups",
                          Ref("GroupNames"),
                          Ref("AWS::NoValue")),
                Users=If("ExternalUsers",
                         Ref("UserNames"),
                         Ref("AWS::NoValue")),
                Condition='CreatePolicy',
            ),
        )
        t.add_resource(
            iam.PolicyType(
                LOGS_POLICY,
                PolicyName='{}-logs'.format(name_prefix),
                PolicyDocument=logs_policy(),
                Roles=If("ExternalRoles",
                         Ref("RoleNames"),
                         Ref("AWS::NoValue")),
                Groups=If("ExternalGroups",
                          Ref("GroupNames"),
                          Ref("AWS::NoValue")),
                Users=If("ExternalUsers",
                         Ref("UserNames"),
                         Ref("AWS::NoValue")),
                Condition='CreatePolicy',
            ),
        )

    def create_template(self):
        self.create_kms_key()
        self.create_policy()
        self.create_bucket()
        self.create_role()



""" The 'base' cloudformation stack that builds out the virtual datacenter.

This includes the VPC, it's subnets, availability zones, etc.
"""

from troposphere import (
    Ref, Output, Join, FindInMap, Select, GetAZs, Not, Equals, Tags, Or,
    Condition, GetAtt, If
)
from troposphere import ec2
from troposphere.route53 import HostedZone, HostedZoneVPCs

from stacker.blueprints.base import Blueprint

NAT_INSTANCE_NAME = 'NatInstance%s'
NAT_GATEWAY_NAME = 'NatGateway%s'
GATEWAY = 'InternetGateway'
GW_ATTACH = 'GatewayAttach'
VPC_NAME = "VPC"
VPC_ID = Ref(VPC_NAME)
DEFAULT_SG = "DefaultSG"
NAT_SG = "NATSG"


class VPC(Blueprint):
    LOCAL_PARAMETERS = {
        "AZCount":  {
            "type": int,
            "default": 2,
        }
    }

    PARAMETERS = {
        "PrivateSubnets": {
            "type": "CommaDelimitedList",
            "description": "Comma separated list of subnets to use for "
                           "non-public hosts. NOTE: Must have as many subnets "
                           "as AZCount"},
        "PublicSubnets": {
            "type": "CommaDelimitedList",
            "description": "Comma separated list of subnets to use for "
                           "public hosts. NOTE: Must have as many subnets "
                           "as AZCount"},
        "InstanceType": {
            "type": "String",
            "description": "NAT EC2 instance type.",
            "default": "m3.medium"},
        "SshKeyName": {
            "type": "AWS::EC2::KeyPair::KeyName"},
        "BaseDomain": {
            "type": "String",
            "default": "",
            "description": "Base domain for the stack."},
        "InternalDomain": {
            "type": "String",
            "default": "",
            "description": "Internal domain name, if you have one."},
        "CidrBlock": {
            "type": "String",
            "description": "Base CIDR block for subnets.",
            "default": "10.128.0.0/16"},
        "ImageName": {
            "type": "String",
            "description": "The image name to use from the AMIMap (usually "
                           "found in the config file.)",
            "default": "NAT"},
        "UseNatGateway": {
            "type": "String",
            "allowed_values": ["true", "false"],
            "description": "If set to true, will configure a NAT Gateway"
                           "instead of NAT instances.",
            "default": "false"},
    }

    def create_conditions(self):
        self.template.add_condition(
            "HasInternalDomain",
            Not(Equals(Ref("InternalDomain"), "")))
        self.template.add_condition(
            "HasExternalDomain",
            Not(Equals(Ref("BaseDomain"), "")))
        self.template.add_condition(
            "HasHostedZones",
            Or(
                Condition("HasInternalDomain"),
                Condition("HasExternalDomain")
            ))
        self.template.add_condition(
            "NoHostedZones",
            Not(Condition("HasHostedZones")))
        self.template.add_condition(
            "UseNatGateway",
            Equals(Ref("UseNatGateway"), "true"))
        self.template.add_condition(
            "UseNatInstances",
            Not(Condition("UseNatGateway")))

    def create_vpc(self):
        t = self.template
        t.add_resource(ec2.VPC(
            VPC_NAME,
            CidrBlock=Ref("CidrBlock"), EnableDnsSupport=True,
            EnableDnsHostnames=True))

        # Just about everything needs this, so storing it on the object
        t.add_output(Output("VpcId", Value=VPC_ID))

    def create_internal_zone(self):
        t = self.template
        t.add_resource(
            HostedZone(
                "InternalZone",
                Name=Ref("InternalDomain"),
                VPCs=[HostedZoneVPCs(
                    VPCId=VPC_ID,
                    VPCRegion=Ref("AWS::Region"))],
                Condition="HasInternalDomain"))
        t.add_output(
            Output(
                "InternalZoneId",
                Value=Ref("InternalZone"),
                Condition="HasInternalDomain"))
        t.add_output(
            Output(
                "InternalZoneName",
                Value=Ref("InternalDomain"),
                Condition="HasInternalDomain"))

    def create_default_security_group(self):
        t = self.template
        t.add_resource(ec2.SecurityGroup(
            DEFAULT_SG,
            VpcId=VPC_ID,
            GroupDescription='Default Security Group'))
        t.add_output(
            Output('DefaultSG',
                   Value=Ref(DEFAULT_SG)))

    def _dhcp_options_hosted_zones(self):
        t = self.template
        domain_name = Join(" ", [Ref("InternalDomain"), Ref("BaseDomain")])
        dhcp_options = t.add_resource(ec2.DHCPOptions(
            'DHCPOptionsWithDNS',
            DomainName=domain_name,
            DomainNameServers=['AmazonProvidedDNS', ],
            Condition="HasHostedZones"))
        t.add_resource(ec2.VPCDHCPOptionsAssociation(
            'DHCPAssociationWithDNS',
            VpcId=VPC_ID,
            DhcpOptionsId=Ref(dhcp_options),
            Condition="HasHostedZones"))

    def _dhcp_options_no_hosted_zones(self):
        t = self.template
        dhcp_options = t.add_resource(ec2.DHCPOptions(
            'DHCPOptionsNoDNS',
            DomainNameServers=['AmazonProvidedDNS', ],
            Condition="NoHostedZones"))
        t.add_resource(ec2.VPCDHCPOptionsAssociation(
            'DHCPAssociationNoDNS',
            VpcId=VPC_ID,
            DhcpOptionsId=Ref(dhcp_options),
            Condition="NoHostedZones"))

    def create_dhcp_options(self):
        self._dhcp_options_hosted_zones()
        self._dhcp_options_no_hosted_zones()

    def create_gateway(self):
        t = self.template
        t.add_resource(ec2.InternetGateway(GATEWAY))
        t.add_resource(ec2.VPCGatewayAttachment(
            GW_ATTACH,
            VpcId=VPC_ID,
            InternetGatewayId=Ref(GATEWAY)))

    def create_network(self):
        t = self.template
        self.create_gateway()
        vpc_id = Ref("VPC")
        t.add_resource(ec2.NetworkAcl('DefaultACL',
                                      VpcId=vpc_id))

        self.create_nat_security_groups()
        subnets = {'public': [], 'private': []}
        net_types = subnets.keys()
        zones = []
        for i in range(self.local_parameters["AZCount"]):
            az = Select(i, GetAZs(""))
            zones.append(az)
            name_suffix = i
            for net_type in net_types:
                name_prefix = net_type.capitalize()
                subnet_name = "%sSubnet%s" % (name_prefix, name_suffix)
                subnets[net_type].append(subnet_name)
                t.add_resource(ec2.Subnet(
                    subnet_name,
                    AvailabilityZone=az,
                    VpcId=vpc_id,
                    DependsOn=GW_ATTACH,
                    CidrBlock=Select(i, Ref("%sSubnets" % name_prefix)),
                    Tags=Tags(type=net_type)))
                route_table_name = "%sRouteTable%s" % (name_prefix,
                                                       name_suffix)
                t.add_resource(ec2.RouteTable(
                    route_table_name,
                    VpcId=vpc_id,
                    Tags=[ec2.Tag('type', net_type)]))
                t.add_resource(ec2.SubnetRouteTableAssociation(
                    "%sRouteTableAssociation%s" % (name_prefix, name_suffix),
                    SubnetId=Ref(subnet_name),
                    RouteTableId=Ref(route_table_name)))

                route_name = '%sRoute%s' % (name_prefix, name_suffix)
                if net_type == 'public':
                    # the public subnets are where the NAT instances live,
                    # so their default route needs to go to the AWS
                    # Internet Gateway
                    t.add_resource(ec2.Route(
                        route_name,
                        RouteTableId=Ref(route_table_name),
                        DestinationCidrBlock="0.0.0.0/0",
                        GatewayId=Ref(GATEWAY)))
                    self.create_nat_instance(i, subnet_name)
                else:
                    # Private subnets are where actual instances will live
                    # so their gateway needs to be through the nat instances
                    t.add_resource(ec2.Route(
                        route_name,
                        RouteTableId=Ref(route_table_name),
                        DestinationCidrBlock='0.0.0.0/0',
                        InstanceId=If(
                            "UseNatInstances",
                            Ref(NAT_INSTANCE_NAME % name_suffix),
                            Ref("AWS::NoValue")),
                        NatGatewayId=If(
                            "UseNatGateway",
                            Ref(NAT_GATEWAY_NAME % name_suffix),
                            Ref("AWS::NoValue"))))

        for net_type in net_types:
            t.add_output(Output(
                "%sSubnets" % net_type.capitalize(),
                Value=Join(",",
                           [Ref(sn) for sn in subnets[net_type]])))
        self.template.add_output(Output(
            "AvailabilityZones",
            Value=Join(",", zones)))

    def create_nat_security_groups(self):
        t = self.template
        # First setup the NAT Security Group Rules
        nat_private_in_all_rule = ec2.SecurityGroupRule(
            IpProtocol='-1', FromPort='-1', ToPort='-1',
            SourceSecurityGroupId=Ref(DEFAULT_SG))

        nat_public_out_all_rule = ec2.SecurityGroupRule(
            IpProtocol='-1', FromPort='-1', ToPort='-1', CidrIp='0.0.0.0/0')

        return t.add_resource(ec2.SecurityGroup(
            NAT_SG,
            VpcId=VPC_ID,
            Condition="UseNatInstances",
            GroupDescription='NAT Instance Security Group',
            SecurityGroupIngress=[nat_private_in_all_rule],
            SecurityGroupEgress=[nat_public_out_all_rule, ]))

    def create_nat_instance(self, zone_id, subnet_name):
        t = self.template
        suffix = zone_id
        nat_instance = t.add_resource(ec2.Instance(
            NAT_INSTANCE_NAME % suffix,
            Condition="UseNatInstances",
            ImageId=FindInMap('AmiMap', Ref("AWS::Region"), Ref("ImageName")),
            SecurityGroupIds=[Ref(DEFAULT_SG), Ref(NAT_SG)],
            SubnetId=Ref(subnet_name),
            InstanceType=Ref('InstanceType'),
            SourceDestCheck=False,
            KeyName=Ref('SshKeyName'),
            Tags=[ec2.Tag('Name', 'nat-gw%s' % suffix)],
            DependsOn=GW_ATTACH))

        eip = t.add_resource(ec2.EIP(
            'NATExternalIp%s' % suffix,
            Domain='vpc',
            InstanceId=If("UseNatInstances",
                          Ref(nat_instance),
                          Ref("AWS::NoValue")),
            DependsOn=GW_ATTACH))

        t.add_resource(ec2.NatGateway(
            NAT_GATEWAY_NAME % suffix,
            Condition="UseNatGateway",
            AllocationId=GetAtt(eip, 'AllocationId'),
            SubnetId=Ref(subnet_name),
        ))

        return nat_instance

    def create_template(self):
        self.create_conditions()
        self.create_vpc()
        self.create_internal_zone()
        self.create_default_security_group()
        self.create_dhcp_options()
        self.create_network()



from troposphere import (
    Ref, ec2, Output, GetAtt, Not, Equals, Condition, And, Join, If
)

from troposphere.elasticache import (
    ReplicationGroup, ParameterGroup, SubnetGroup
)

from troposphere.route53 import RecordSetType

from stacker.blueprints.base import Blueprint

# Resource name constants
SUBNET_GROUP = "SubnetGroup"
SECURITY_GROUP = "SecurityGroup"
REPLICATION_GROUP = "ReplicationGroup"
DNS_RECORD = "ReplicationGroupDnsRecord"
PARAMETER_GROUP = "ParameterGroup"


class BaseReplicationGroup(Blueprint):
    """Base Blueprint for all Elasticache ReplicationGroup blueprints.

    ReplicationGroups are only currently supported by the redis engine.
    """

    ALLOWED_ENGINES = ["redis"]

    LOCAL_PARAMETERS = {
        "ClusterParameters": {
            "type": dict,
            "default": {},
        },
    }

    def engine(self):
        return None

    def get_engine_versions(self):
        """Used by engine specific subclasses - returns valid engine versions.

        Should only be overridden if the class variable ENGINE is defined on
        the class.

        Return:
            list: A list of valid engine versions for the given engine.
        """
        return []

    def get_parameter_group_family(self):
        """Used by engine specific subclasses to return parameter group family.

        Should only be overridden if the class variable ENGINE is defined on
        the class.

        Return:
            list: A list of valid parameter group families for the given
                  engine.
        """
        return []

    def _get_parameters(self):
        parameters = {
            "VpcId": {
                "type": "AWS::EC2::VPC::Id",
                "description": "Vpc Id to place the Cluster in"
            },
            "Subnets": {
                "type": "List<AWS::EC2::Subnet::Id>",
                "description": "Subnets to deploy the Cluster nodes in."
            },
            "ParameterGroupFamily": {
                "type": "String",
                "description": "The parametergroup family to use, dependent "
                               "on the engine.",
                "allowed_values": self.get_parameter_group_family()
            },
            "AutomaticFailoverEnabled": {
                "type": "String",
                "description": "Specifies whether a read-only replica will be "
                               "automatically promoted to read/write primary "
                               "if the existing primary fails. If true, "
                               "Multi-AZ is enabled for this replication "
                               "group. If false, Multi-AZ is disabled for "
                               "this replication group. If true, "
                               "NumCacheClusters must be at least 2.",
                "default": "true",
                "allowed_values": ["true", "false"]
            },
            "AutoMinorVersionUpgrade": {
                "type": "String",
                "description": "Set to 'true' to allow minor version upgrades "
                               "during maintenance windows.",
                "default": "false",
                "allowed_values": ["true", "false"]
            },
            "CacheNodeType": {
                "type": "String",
                "description": "AWS ElastiCache Cache Node Type",
                "default": "cache.t2.medium"
            },
            "EngineVersion": {
                "type": "String",
                "description": "Engine version for the Cache Cluster.",
            },
            "NotificationTopicArn": {
                "type": "String",
                "description": "ARN of the SNS Topic to publish events to.",
                "default": "",
            },
            "NumCacheClusters": {
                "type": "Number",
                "description": "The number of cache clusters this replication "
                               "group will initially have. If Multi-AZ "
                               "(ie: the AutomaticFailoverEnabled Parameter) "
                               "is enabled, the value of this parameter must "
                               "be at least 2.",
                "default": "2",
                "min_value": "1",
            },
            "Port": {
                "type": "Number",
                "description": "The port to run the cluster on.",
                "default": "0",
            },
            "PreferredCacheClusterAZs": {
                "type": "CommaDelimitedList",
                "description": "Must match the # of nodes in "
                               "NumCacheClusters.",
                "default": "",
            },
            "PreferredMaintenanceWindow": {
                "type": "String",
                "description": "A (minimum 60 minute) window in "
                               "DDD:HH:MM-DDD:HH:MM format in UTC for "
                               "backups. Default: Sunday 3am-4am PST",
                "default": "Sun:11:00-Sun:12:00"
            },
            "SnapshotArns": {
                "type": "CommaDelimitedList",
                "description": "A list of s3 ARNS where redis snapshots are "
                               "stored that will be used to create the "
                               "cluster.",
                "default": "",
            },
            "SnapshotRetentionLimit": {
                "type": "Number",
                "description": "The number of daily snapshots to retain. Only "
                               "valid for clusters with the redis Engine.",
                "default": "0",
            },
            "SnapshotWindow": {
                "type": "String",
                "description": "For Redis cache clusters, daily time range "
                               "(in UTC) during which ElastiCache will begin "
                               "taking a daily snapshot of your node group. "
                               "For example, you can specify 05:00-09:00.",
                "default": ""
            },
            "InternalZoneId": {
                "type": "String",
                "default": "",
                "description": "Internal zone Id, if you have one."
            },
            "InternalZoneName": {
                "type": "String",
                "default": "",
                "description": "Internal zone name, if you have one."
            },
            "InternalHostname": {
                "type": "String",
                "default": "",
                "description": "Internal domain name, if you have one."
            },
        }

        engine_versions = self.get_engine_versions()
        if engine_versions:
            parameters['EngineVersion']['allowed_values'] = engine_versions

        if not self.engine():
            parameters['Engine'] = {
                "type": "String",
                "description": "Database engine for the RDS Instance.",
                "allowed_values": self.ALLOWED_ENGINES
            }
        else:
            if self.engine() not in self.ALLOWED_ENGINES:
                raise ValueError("ENGINE must be one of: %s" %
                                 ", ".join(self.ALLOWED_ENGINES))

        return parameters

    def create_conditions(self):
        t = self.template

        t.add_condition(
            "DefinedNotificationArn",
            Not(Equals(Ref("NotificationTopicArn"), "")))
        t.add_condition(
            "DefinedPort",
            Not(Equals(Ref("Port"), "0")))
        t.add_condition(
            "DefinedAvailabilityZones",
            Not(Equals(Join(",", Ref("PreferredCacheClusterAZs")), "")))
        t.add_condition(
            "DefinedSnapshotArns",
            Not(Equals(Join(",", Ref("SnapshotArns")), "")))
        t.add_condition(
            "DefinedSnapshotWindow",
            Not(Equals(Ref("SnapshotWindow"), "")))

        # DNS Conditions
        t.add_condition(
            "HasInternalZone",
            Not(Equals(Ref("InternalZoneId"), "")))
        t.add_condition(
            "HasInternalZoneName",
            Not(Equals(Ref("InternalZoneName"), "")))
        t.add_condition(
            "HasInternalHostname",
            Not(Equals(Ref("InternalHostname"), "")))
        t.add_condition(
            "CreateInternalHostname",
            And(Condition("HasInternalZone"),
                Condition("HasInternalZoneName"),
                Condition("HasInternalHostname")))

    def create_parameter_group(self):
        t = self.template
        params = self.local_parameters["ClusterParameters"]
        t.add_resource(
            ParameterGroup(
                PARAMETER_GROUP,
                Description=self.name,
                CacheParameterGroupFamily=Ref("ParameterGroupFamily"),
                Properties=params,
            )
        )

    def create_subnet_group(self):
        t = self.template
        t.add_resource(
            SubnetGroup(
                SUBNET_GROUP,
                Description="%s subnet group." % self.name,
                SubnetIds=Ref("Subnets")))

    def create_security_group(self):
        t = self.template
        sg = t.add_resource(
            ec2.SecurityGroup(
                SECURITY_GROUP,
                GroupDescription="%s security group" % self.name,
                VpcId=Ref("VpcId")))
        t.add_output(Output("SecurityGroup", Value=Ref(sg)))

    def create_replication_group(self):
        t = self.template
        availability_zones = If("DefinedAvailabilityZones",
                                Ref("PreferredCacheClusterAZs"),
                                Ref("AWS::NoValue"))
        t.add_resource(
            ReplicationGroup(
                REPLICATION_GROUP,
                AutomaticFailoverEnabled=Ref("AutomaticFailoverEnabled"),
                AutoMinorVersionUpgrade=Ref("AutoMinorVersionUpgrade"),
                CacheNodeType=Ref("CacheNodeType"),
                CacheParameterGroupName=Ref(PARAMETER_GROUP),
                CacheSubnetGroupName=Ref(SUBNET_GROUP),
                Engine=self.engine() or Ref("Engine"),
                EngineVersion=Ref("EngineVersion"),
                NotificationTopicArn=If("DefinedNotificationArn",
                                        Ref("NotificationTopicArn"),
                                        Ref("AWS::NoValue")),
                NumCacheClusters=Ref("NumCacheClusters"),
                Port=If("DefinedPort",
                        Ref("Port"),
                        Ref("AWS::NoValue")),
                PreferredCacheClusterAZs=availability_zones,
                PreferredMaintenanceWindow=Ref("PreferredMaintenanceWindow"),
                ReplicationGroupDescription=self.name,
                SecurityGroupIds=[Ref(SECURITY_GROUP), ],
                SnapshotArns=If("DefinedSnapshotArns",
                                Ref("SnapshotArns"),
                                Ref("AWS::NoValue")),
                SnapshotRetentionLimit=Ref("SnapshotRetentionLimit"),
                SnapshotWindow=If("DefinedSnapshotWindow",
                                  Ref("SnapshotWindow"),
                                  Ref("AWS::NoValue")),
            )
        )

    def get_primary_address(self):
        return GetAtt(REPLICATION_GROUP, "PrimaryEndPoint.Address")

    def get_secondary_addresses(self):
        return GetAtt(REPLICATION_GROUP, "ReadEndPoint.Addresses.List")

    def create_dns_records(self):
        t = self.template
        primary_endpoint = self.get_primary_address()

        t.add_resource(
            RecordSetType(
                DNS_RECORD,
                HostedZoneId=Ref("InternalZoneId"),
                Comment="ReplicationGroup CNAME Record",
                Name=Join(".", [Ref("InternalHostname"),
                          Ref("InternalZoneName")]),
                Type="CNAME",
                TTL="120",
                ResourceRecords=[primary_endpoint],
                Condition="CreateInternalHostname"))

    def create_cluster_outputs(self):
        t = self.template
        t.add_output(Output("PrimaryAddress",
                            Value=self.get_primary_address()))
        t.add_output(Output("ReadAddresses",
                            Value=Join(",", self.get_secondary_addresses())))

        t.add_output(Output("ClusterPort",
                            Value=GetAtt(REPLICATION_GROUP,
                                         "PrimaryEndPoint.Port")))
        t.add_output(Output("ClusterId", Value=Ref(REPLICATION_GROUP)))
        t.add_output(
            Output(
                "PrimaryCname",
                Condition="CreateInternalHostname",
                Value=Ref(DNS_RECORD)))

    def create_template(self):
        self.create_conditions()
        self.create_parameter_group()
        self.create_subnet_group()
        self.create_security_group()
        self.create_replication_group()
        self.create_dns_records()
        self.create_cluster_outputs()



from stacker.blueprints.base import Blueprint

from troposphere import (
    dynamodb2,
    Ref,
    GetAtt,
    Output,
)

from . import util

MAX_GSI_VALUE = 5
MAX_LSI_VALUE = 5


def prep_schemata(config):
    schemata = []
    try:
        schemas = config["KeySchema"]
    except KeyError:
        raise KeyError(
            "The key schema is required for the creation of a DynamoDB table."
        )

    for schema in schemas:
        schemata.append(dynamodb2.KeySchema(**schema))

    return schemata


def prep_projection(config):
    return dynamodb2.Projection(**config["Projection"])


def prep_throughput(config):
    try:
        return dynamodb2.ProvisionedThroughput(
            **config["ProvisionedThroughput"]
        )
    except KeyError:
        raise KeyError(
            "ProvisionedThroughput values are required for the creation of " +
            "a DynamoDB table or index."
        )


def prep_config(raw_config):
    prepped_config = {}

    dynamodb_table_properties = [
        "AttributeDefinitions",
        "GlobalSecondaryIndexes",
        "KeySchema",
        "LocalSecondaryIndexes",
        "ProvisionedThroughput",
        "StreamSpecification",
    ]

    util.check_properties(
        raw_config,
        dynamodb_table_properties,
        "DynamoDB",
    )

    # AttributeDefinitions are required, so raise a KeyError if this doesn't
    # work.
    try:
        config_attributes = raw_config["AttributeDefinitions"]
    except KeyError:
        raise KeyError(
            "Attribute definitions are required for the creation of a " +
            "DynamoDB table."
        )

    attributes = []
    for attribute in config_attributes:
        attributes.append(dynamodb2.AttributeDefinition(**attribute))

    prepped_config["AttributeDefinitions"] = attributes

    if "GlobalSecondaryIndexes" in raw_config:
        # AWS limits us to 5 GSIs.  Check for that and bail if there's more.
        if len(raw_config["GlobalSecondaryIndexes"]) > MAX_GSI_VALUE:
            raise ValueError(
                "A DynamoDB table can only have a maximum of 5 global " +
                "secondary indexes."
            )

        gsis = []
        for gsi in raw_config["GlobalSecondaryIndexes"]:
            gsi["KeySchema"] = prep_schemata(gsi)
            gsi["Projection"] = prep_projection(gsi)
            gsi["ProvisionedThroughput"] = prep_throughput(gsi)
            gsis.append(dynamodb2.GlobalSecondaryIndex(**gsi))

        prepped_config["GlobalSecondaryIndexes"] = gsis

    prepped_config["KeySchema"] = prep_schemata(raw_config)

    if "LocalSecondaryIndexes" in raw_config:
        # Another limit of 5.  Check and bail if more than that.
        if len(raw_config["LocalSecondaryIndexes"]) > MAX_LSI_VALUE:
            raise ValueError(
                "A DynamoDB table can only have a maximum of 5 local " +
                "secondary indexes."
            )

        lsis = []
        for lsi in raw_config["LocalSecondaryIndexes"]:
            lsi["KeySchema"] = prep_schemata(lsi)
            lsi["Projection"] = prep_projection(lsi)
            lsi["ProvisionedThroughput"] = prep_throughput(lsi)
            lsis.append(dynamodb2.LocalSecondaryIndex(**lsi))

        prepped_config["LocalSecondaryIndexes"] = lsis

    prepped_config["ProvisionedThroughput"] = prep_throughput(raw_config)

    if "StreamSpecification" in raw_config:
        prepped_config["StreamSpecification"] = dynamodb2.StreamSpecification(
            **raw_config["StreamSpecification"]
        )

    return prepped_config


def validate_tables(tables):
    prepped_configs = {}
    for table_name, table_config in tables.iteritems():
        prepped_config = prep_config(table_config)
        prepped_configs[table_name] = prepped_config

    return prepped_configs


class DynamoDB(Blueprint):
    """Manages the creation of DynamoDB tables."""

    VARIABLES = {
        "Tables": {
            "type": dict,
            "description": "Dictionary of DynamoDB table definitions",
            "validator": validate_tables,
        }
    }

    def create_template(self):
        variables = self.get_variables()

        for table_name, table_config in variables["Tables"].iteritems():
            stream_enabled = "StreamSpecification" in table_config
            self.create_table(table_name, table_config, stream_enabled)

    def create_table(self, table_name, table_config, stream_enabled):
        t = self.template

        t.add_resource(
            dynamodb2.Table(
                table_name,
                **table_config
            )
        )

        if stream_enabled:
            t.add_output(
                Output(table_name + "StreamArn",
                       Value=GetAtt(table_name, "StreamArn")))

        t.add_output(Output(table_name + "Name", Value=Ref(table_name)))
